/* SiME File Events code */

/* Copyright 2021 Kazuo K.
 * SiME is licensed under the MIT license. You should have
 * received a copy of the license with your distribution.
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>

#include <Xm/Xm.h>
#include <Xm/FileSB.h>
#include <Xm/Text.h>

#include "sys_consts.h"
#include "globals.h"
#include "utils.h"
#include "undo.h"
#include "callbacks.h"
#include "fileevents.h"
#include "optionsevents.h"
#include "heap_string.h"

static void    set_filename(char *filename);
static void    gui_createFileDialog(Widget parent, FileMode mode);
static void    gui_cancelBrowseCB(Widget w, XtPointer client_data, XtPointer call_data);
static void    gui_browseFileCB(Widget w, XtPointer client_data, XtPointer call_data);
static void    AutoWW(char *name);
static void    gui_saveFileCB(Widget w, XtPointer client_data, XtPointer call_data);
static Boolean gui_save(char *filename, FileMode mode);
static Boolean savefile(char *filename, FileMode mode);

void auto_save_func(XtPointer client_data, XtIntervalId *id)
{
   hstring  filename;
   char    *home, *name;

   set_cursor(toplevel, TimeCursor);
   set_cursor(text_canvas, TimeCursor);
   set_status_message("Please wait: Autosaving");

   name = strrchr(FileName, '/');
   if (name) name++;
   else      name = FileName;

   home = getenv("HOME");
   if (!home) home = "";

   heap_string_compose(&filename, home, sime_path, "/", name, NULL);
   if (!savefile(filename.str, SAVE_ALL))
      gui_error("Sorry unable to perform autosave");
   heap_string_free(&filename);

   set_status_message("");
   unset_cursor(toplevel);
   unset_cursor(text_canvas);

   if (AutoSave)
      XtAppAddTimeOut(app_context, AutoSaveTime*1000, auto_save_func, client_data);
}

void  gui_browse_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
   gui_createFileDialog(toplevel, (uintptr_t)client_data);
}


Boolean  loadfile(char *filename, FileMode mode)
{
   int      stdin_flag = !strcmp(filename, "-");
   struct stat stat_val;
   char    *filecontents;
   FILE    *fpointer;
   int      akt_len;
   int      f_length;

   if (stdin_flag || !stat(filename, &stat_val))
   {
      if (stdin_flag)
      {
         fpointer = stdin;
         f_length = 1024; // Block length to read
      }
      else
      {
         f_length = stat_val.st_size;
         fpointer = fopen(filename, "r");
      }
      filecontents = NULL;
      akt_len = 0;
      if (fpointer != NULL)
      {
	 while (f_length && !feof(fpointer))
	 {
	    filecontents = XtRealloc(filecontents, (unsigned)(akt_len + f_length + 1));
	    akt_len += fread(filecontents+akt_len, sizeof(char), f_length, fpointer);
	    filecontents[akt_len] = '\0';
	 }
	 if (!stdin_flag)
	    fclose(fpointer);

	 if (mode == OPENIT)
	 {
	    undo_redo_action = True;
	    if (filecontents)
	       XmTextSetString(text_canvas, filecontents);
	    else
	       XmTextSetString(text_canvas, "");
	    reset_undo_redo();
	    AutoWW(filename);
	 }
	 else if (mode == APPENDIT)
	 {
	    if (filecontents)
	       XmTextInsert(text_canvas, XmTextGetLastPosition(text_canvas), filecontents);
	 }
	 else
	 {
	    if (filecontents)
	       XmTextInsert(text_canvas, XmTextGetInsertionPosition(text_canvas), filecontents);
	 }
	 XtFree(filecontents);
	 return(True);
      }
      else
	 return(False);
   }
   else
   {
      if (errno == ENOENT)
      {
         AutoWW(filename);
      }
      return False;
   }
}

static void set_filename(char *filename)
{
   XtFree(FileName);
   FileName = filename;
   set_title_string();
}

static void  gui_createFileDialog(Widget parent, FileMode mode)
{
   Widget  fileD;
   Arg     args[5];
   int     n = 0;
   char   *w_name = NULL;

   switch(mode)
   {
      case APPENDIT:
      case OPENIT :
         w_name = "Open File";
         break;

      case SAVE_AS:
      case SAVE_ALL:
         w_name = "saveFileD";
         break;

   }

   XtSetArg(args[n], XmNdeleteResponse, XmDESTROY); n++;
   fileD = XmCreateFileSelectionDialog(parent, w_name, args, n);

   if (mode == OPENIT) 
      XtAddCallback(fileD, XmNokCallback, gui_browseFileCB, (XtPointer)mode);
   else
      XtAddCallback(fileD, XmNokCallback, gui_saveFileCB, (XtPointer)mode);

   XtAddCallback(fileD, XmNcancelCallback, gui_cancelBrowseCB, (XtPointer)mode);

   XtUnmanageChild(XmFileSelectionBoxGetChild(fileD, XmDIALOG_HELP_BUTTON));

   XtManageChild(fileD);
}

static void  gui_cancelBrowseCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   XtDestroyWidget(w);
}

static void  gui_browseFileCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   XmFileSelectionBoxCallbackStruct *cbd =
		(XmFileSelectionBoxCallbackStruct *)call_data;
   FileMode mode = (FileMode)client_data;
   char    *filename = NULL;
   char     answer;

   /* Get the filename from the file selection box */
   XmStringGetLtoR(cbd->value, CHARSET, &filename);

   /* Make sure the user doesn't lose data accidentally */
   if (mode == OPENIT)
   {
      if (changes_counter)
	 answer = gui_query("Unsaved Text. Are you sure about this?");
      else
         answer = YES;

      if (answer == SAVE_THEN_DO)
         gui_save_CB(w, NULL, NULL);

      if ((answer == YES) || (answer == SAVE_THEN_DO))
      {
         set_status_message("Attempting to open file");
	 set_cursor(w, TimeCursor);
	 if (loadfile(filename, mode))
	 {
	    if (FileBox)
	       XtDestroyWidget(w);
	    set_filename(filename);
	    XmTextSetCursorPosition(text_canvas, 0);
            set_status_message("");
	 }
	 else
	 {
	    set_status_message("Sorry! Unable to open file");
	 }
	 unset_cursor(w);
      }
   }
   else
   {
      set_status_message("Attempting to insert file");
      set_cursor(w, TimeCursor);
      if (loadfile(filename, mode))
      {
	    if (FileBox)
	       XtDestroyWidget(w);
            set_status_message("");
      }
      else
	 set_status_message("Sorry! Unable to open file");
      unset_cursor(w);
   }
}


/* Automatically determine the wordwrap setting from filename */
static void AutoWW(char *name)
{
   int    i;
   char  *p;

   for (i=0; i<extnum; i++)
   {
      p = strrchr(name, '.');  // Look at the file extension		
      if (!p)
         continue;

      ++p;		
      if (!strcmp(p, extmode[i].ext))
      {
         WordWrap   = extmode[i].ww;  // load wordwrap
         AutoIndent = extmode[i].ai;  // load autoindent
         set_autoindent_wordwrap();  // set them using this function
         return;
      }
   }
}




/* Save the contents of the text widget to file */
void  gui_saveas_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
   gui_createFileDialog(toplevel, SAVE_AS);
}


/* Save the contents of the text widget to file */
void  gui_save_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
   if ( (strcmp(FileName, "") == 0) ||
        (strcmp(FileName, "-") == 0) ||
        (strcmp(FileName, "untitled") == 0) )
   {
      gui_createFileDialog(toplevel, SAVE_ALL);
   }
   else
   {
      if (gui_save(FileName, SAVE_ALL))
      {
#ifdef PERSISTENT_UNDO
         clear_undo_redo();
#else
	 reset_undo_redo();
#endif
      }
   }
}

static void  gui_saveFileCB(Widget w, XtPointer client_data, XtPointer call_data)
{
   XmFileSelectionBoxCallbackStruct *cbd =
		(XmFileSelectionBoxCallbackStruct *)call_data;
   FileMode mode = (FileMode)client_data;
   char    *filename = NULL;

   /* Get the filename from the file selection box */
   XmStringGetLtoR(cbd->value, CHARSET, &filename);

   if (gui_save(filename, (int)mode))
   {
      set_filename(filename);
      if (mode == SAVE_ALL)
      {
#ifdef PERSISITENT_UNDO
         clear_undo_redo();
#else
	 reset_undo_redo();
#endif
      }
      else if (mode == SAVE_AS)
         reset_undo_redo();

      if (FileBox)
	 XtDestroyWidget(w);
   }
}

static Boolean gui_save(char *filename, FileMode mode)
{
   Boolean  flag;

   set_status_message("Attempting to save file");
   set_cursor(toplevel, TimeCursor);
   set_cursor(text_canvas, TimeCursor);

   flag = savefile(filename, mode);
   if (flag)
      set_status_message("Ok file saved");
   else if (gui_error("Sorry! Unable to save file ") == HELP)
      gui_info("You may not have specified a valid file name.\n"
	       "Or you may not have write permission to that file.");

   set_cursor(toplevel, MainCursor);
   set_cursor(text_canvas, TextCursor);

   return(flag);
}

static Boolean savefile(char *filename, FileMode mode)
{
   Boolean  flag = FALSE;
   char    *textcontents;
   FILE    *filepointer;

   if (*filename)
   {
      if (AutoBackup && (access(filename, F_OK) == 0))
      {
         hstring  backup;
         heap_string_compose(&backup, filename, ".backup", NULL);
	 rename(filename, backup.str);
	 heap_string_free(&backup);
      }
      if ((filepointer = fopen(filename, "w+")) != NULL)
      {
	 /* Get the contents of the text widget text_canvas. */
	    textcontents = XmTextGetString(text_canvas);
	 if (textcontents != NULL)
	 {
	    int  len = strlen(textcontents);

	    /* Write out the file */
	    fwrite(textcontents, sizeof(char), len, filepointer);
	    if (len && (textcontents[len - 1] != '\n'))
	       fputc('\n', filepointer);
 	    XtFree(textcontents);

	    flag = TRUE;
	 }
	 fclose(filepointer);
	 sync();
      }
   }
   return(flag);
}


void  gui_new_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
   char     answer;

   /* Check to see if text has changed first before wiping it out */
   if (changes_counter)
      answer = gui_query("Unsaved Text. Are you sure about this?");
   else
      answer = YES;

   if (answer == SAVE_THEN_DO)
      gui_save_CB(w, NULL, NULL);

   if ((answer == YES) || (answer == SAVE_THEN_DO))
   {
      /* Now blank out the text_canvas and the file name fields */
      set_filename(Xstrdup("untitled"));
      XmTextSetString(text_canvas, "");
      reset_undo_redo();
      SelectedOnly = False;
   }
}


void  gui_quit_CB(Widget w, XtPointer client_data, XtPointer call_data)
{
   char  prompt;

   /* Confirm if user wants to close an unsaved file */
   if (changes_counter)
   {
      prompt = gui_query("The file open is unsaved.\n"
	  "Do you wish to save it before quitting?");

      if (prompt == SAVE_THEN_DO)
	 gui_save_CB(w, NULL, NULL);

      if (prompt == NO)
         return;
   }
   XtFree((char*)extmode);
   XtFree(FileName);
   exit(0);
}
